/**
 *
 * Compiz galaxy livewallpaper plugin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2011 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 *
 * Several parts taken from the original plugin developed by Pal Dorogi
 * with the following copyright:
 *
 * Copyright (C) 2010 Pal Dorogi <pal.dorogi@gmail.com>
 *
 * Original plugin can be found here:
 * http://launchpad.net/compiz-plugins-livewallpapers
 *
 **/

#include <X11/Xatom.h>
#include <X11/extensions/Xrender.h>
#include <X11/extensions/shape.h>

#include <core/core.h>
#include <core/pluginclasshandler.h>
#include <core/atoms.h>

#include <composite/composite.h>
#include <opengl/opengl.h>

#include "particle.h"
#include "galaxy_options.h"

class GalaxyScreen:
	public ScreenInterface,
	public CompositeScreenInterface,
	public GLScreenInterface,
	public PluginClassHandler<GalaxyScreen, CompScreen>,
	public GalaxyOptions
{
	public:
		/**************
		 * Galaxy state
		 **************/
		enum GalaxyState {
			/*
			 * State will be undefined if an error occurs, for example if an bad_alloc
			 * exception is catched. It prevents the galaxy from beeing activated again
			 * until the problem is solved (with the help of the user).
			 */
			GalaxyUndefined,

			/*
			 * State will be true if the galaxy is active and the animation should be
			 * updated.
			 */
			GalaxyTrue,

			/*
			 * State will be false if the galaxy is paused or deactivated, so the galaxy
			 * won't be updated and the original desktop will be shown (if available).
			 */
			GalaxyFalse
		};

		/****************************
		 * Constructor and Destructor
		 ****************************/
		GalaxyScreen(CompScreen *);
		~GalaxyScreen();

		/*******************
		 * Public attributes
		 *******************/
		CompositeScreen *cScreen;
		GLScreen *gScreen;

		/* The fake desktop window. */
		Window fakeDesktop;

		/* The galaxy animation will be rendered to this texture */
		GLTexture::List galaxyTexture;
		CompSize galaxySize;

		/* The state of the galaxy */
		GalaxyState state;

		/*****************************
		 * Fake desktop window methods
		 *****************************/
		void createFakeDesktopWindow();
		void destroyFakeDesktopWindow();

		/*****************
		 * Drawing methods
		 *****************/
		/* Draws the background texture. */
		void drawSpace();

		/* Draws the light texture to a plane next to the stars. */
		void drawLight();

		/* Draws the galaxy to the current opengl buffer. */
		void drawGalaxy();

		/*
		 * Draws the galaxy to the texture using the framebuffer extension if
		 * available.
		 */
		void drawGalaxyToTexture();

		/**************
		 * Screen hooks
		 **************/
		/* Updates each particle. */
		void preparePaint(int);

		/* Updates the galaxy texture. */
		bool glPaintOutput(const GLScreenPaintAttrib&, const GLMatrix&,
				const CompRegion&, CompOutput*, unsigned int);

		/* Triggers repainting. / Damage the whole screen. */
		void donePaint();

		/* Handles creation of the fake desktop window. */
		void handleEvent(XEvent*);

		/**********************
		 * Toggle key callbacks
		 **********************/
		bool initGalaxy(CompAction*, CompAction::State, CompOption::Vector&);
		bool termGalaxy(CompAction*, CompAction::State, CompOption::Vector&);

		/*************************
		 * Option handling methods
		 *************************/
		/* Changes params of the particle system. */
		void particlesOptionChanged(CompOption*, Options);

		/* Updates the matrix on changes. */
		void matrixOptionChanged(CompOption*, Options);

	private:
		/********************
		 * Private attributes
		 ********************/
		/* Load and store the light texture. */
		GLTexture::List lightTexture;
		CompSize lightSize;
		void loadLightTexture();

		/* Load and store the space background texture. */
		GLTexture::List spaceTexture;
		CompSize spaceSize;
		GLMatrix spaceMatrix;
		void loadSpaceTexture();

		/* If available it represents an opengl framebuffer object. */
		GLuint galaxyBuffer;

		/* Stores the current transformation matrix of the galaxy. */
		GLMatrix matrix;
		void updateMatrix();

		/* The particle system handling all stars. */
		SpiralGalaxy *ps;

		/*
		 * _COMPIZ_WALLPAPER_SUPPORTED atom is used to indicate that
		 * the wallpaper plugin or a plugin providing similar functionality is
		 * active so that desktop managers can respond appropriately.
		 * Taken from compiz's wallpaper plugin.
		 */
		Atom compizWallpaperAtom;
};

class GalaxyWindow:
	public PluginClassHandler<GalaxyWindow, CompWindow>,
	public CompositeWindowInterface,
	public GLWindowInterface
{
	public:
		GalaxyWindow(CompWindow *);

		/*******************
		 * Public attributes
		 *******************/
		CompWindow *window;
		CompositeWindow *cWindow;
		GLWindow *gWindow;

		/*
		 * Draws the GalaxyScreen->galaxyTexture on a fullscreen plane. The original
		 * background won't be drawn if galaxy is active. The fake desktop window
		 * gets always drawn if available.
		 */
		bool glDraw (const GLMatrix &, GLFragment::Attrib &,
				const CompRegion &, unsigned int);
};


#define GALAXY_SCREEN(s) \
	GalaxyScreen *gs = GalaxyScreen::get(s)

#define GALAXY_WINDOW(w) \
	GalaxyWindow *gw = GalaxyWindow::get(w)


class GalaxyPluginVTable:
	public CompPlugin::VTableForScreenAndWindow<GalaxyScreen, GalaxyWindow>
{
	public:
		bool init();
};
